indexing

	description:

		"Parsing that can do streaming validation on events."

	implementation:
	"For a streaming parser there are just three events: %
	%   a. start_tag or%
	%   b. start_attribute.%
	%   c. content (tag or attribute)%
	%This parser works on the principle if such an event is allowed, %
	%and if it is, if it is of the correct type. For tag/attribute the name%
	%should be ok, and for data it should be of the correct type."

	known_issues:
	"1. Should be able to set allowed roots. You cannot in DTD however."

	library: "Gobo Eiffel XML Library"
	copyright: "Copyright (c) 2002, Berend de Boer and others"
	license: "Eiffel Forum License v1 (see forum.txt)"
	date: "$Date: $"
	revision: "$Revision: $"

class

	XM_VALIDATING_PARSER


inherit

	XM_PARSER
		redefine
			last_error_extended_description
		end

	XM_FORWARD_CALLBACKS
		redefine
			on_attribute,
			on_content,
			on_end_tag,
			on_error,
			on_finish,
			on_start,
			on_start_tag,
			on_start_tag_finish
		end

	XM_FORWARD_DTD_CALLBACKS
		redefine
			on_doctype
		end

	XM_HELPER_ROUTINES
		export
			{NONE} all
		end

	GF_FORMAT


creation

	make,
	make_with_resolver


feature {NONE} -- Initialization

	make (a_feed: XM_PARSER) is
			-- Initialize. `a_feed' should not resolve name spaces.
		require
			feed_not_void: a_feed /= Void
		local
			a_resolver: XM_NAMESPACE_RESOLVER
		do
			create a_resolver.set_next (Current)
			make_with_resolver (a_feed, a_resolver)
		end

	make_with_resolver (a_feed: XM_PARSER; a_resolver: XM_NAMESPACE_RESOLVER) is
			-- Initialize. `a_feed' should not resolve name spaces.
		require
			feed_not_void: a_feed /= Void
			resolver_not_void: a_resolver /= Void
		do
			is_valid := True
			namespace_resolver := a_resolver
			namespace_resolver.set_next (Current)
			feed := a_feed
			feed.set_callbacks (namespace_resolver)
			feed.set_dtd_callbacks (Current)
			create elements.make (1024) -- he, we're not written for trivial DTDs!

			-- Resolvers are not applicable I think.
			create {XM_NULL_EXTERNAL_RESOLVER} dtd_resolver
			entity_resolver := dtd_resolver

			-- We forward callbacks, make sure to stop somewhere.
			create {XM_CALLBACKS_NULL} callbacks.make

			cached_content := ""
		end


feature -- Parsing

	parse_from_stream (a_stream: KI_CHARACTER_INPUT_STREAM) is
			-- Parse XML document from input stream.
		do
			feed.parse_from_stream (a_stream)
		end

	parse_from_string (a_string: STRING) is
			-- Parse XML document from `a_string'.
		do
			feed.parse_from_string (a_string)
		end


feature -- Incremental parsing

	parse_incremental_from_stream (a_stream: KI_CHARACTER_INPUT_STREAM) is
			-- Parse partial XML document from input stream.
			-- After the last part of the data has been fed into the parser,
			-- call `set_end_of_document' to get any pending error messages.
		do
			feed.parse_incremental_from_stream (a_stream)
		end

	parse_incremental_from_string (a_data: STRING) is
			-- Parse partial XML document from 'a_data'.
			-- Note: You can call `parse_incremental_from_string' multiple
			-- times and give the parse the document in parts only.
			-- You have to call `set_end_of_file' after the last call to
			-- 'parse_incremental_from_string' in every case.
		do
			feed.parse_incremental_from_string (a_data)
		end

	finish_incremental is
			-- Call this routine to tell the parser that the document
			-- has been completely parsed and no input is coming anymore.
		do
			feed.finish_incremental
		end


feature -- Status report

	is_incremental: BOOLEAN is
			-- Can parser handle incremental input? If yes, you can feed
			-- the parser with a document in several steps. You must use
			-- the special parsing routines (the ones that contain
			-- "incremental" in their name) to do this and call
			-- `set_end_of_document' after the last part has been fed.
		do
			Result := feed.is_incremental
		end


feature -- Access

	source: XM_SOURCE is
			-- Source of the XML document being parsed.
		do
			Result := feed.source
		end

	position: XM_POSITION is
			-- Current position in the source of the XML document
		do
			Result := feed.position
		end


feature -- Error reporting

	is_correct: BOOLEAN is
			-- Has no error been detected?
		do
			Result := is_valid and then feed.is_correct
		end

	last_error: INTEGER is
			-- Code of last error
			-- (See XM_ERROR_CODES.)
		do
			Result := feed.last_error
		end

	last_error_description: STRING is
			-- Textual description of last error.
		do
			Result := feed.last_error_description
		end

	last_error_extended_description: STRING is
			-- Verbose textual description of last error.
		do
			if validation_error_extended_description = Void then
				Result := feed.last_error_extended_description
			else
				Result := validation_error_extended_description
			end
		end

	last_error_position: XM_POSITION is
			-- Attempt to give an accurate position of the last (validation) error
		do
			if validation_error_position = Void then
				Result := position
			else
				Result := validation_error_position
			end
		end


feature {NONE} -- Document

	on_start is
			-- Forward start.
		local
			xp: XM_EXPAT_PARSER
		do
			create state.make
			is_valid := True
			validation_error_extended_description := Void
			validation_error_position := Void
			cached_content.wipe_out
			precursor

			-- Close your eyes, awful hack follows
			xp ?= feed
			if xp /= Void then
				-- We might have unregistered stuff because of an error,
				-- so register the default stuff again. What if the user
				-- has requested more callbacks? Yes, you're screwed.
				xp.register_default_callbacks
				xp.register_doctype_handler
				-- Why do I need parameter entity parsing?? Where used??
				-- xp.enable_parameter_entity_parsing
			end

		ensure
			have_state: state /= Void
			cached_content_empty: cached_content.count = 0
		end

	on_finish is
			-- Forward finish.
		do
			cached_content.wipe_out
			precursor
		end


feature {NONE} -- Document type definition callbacks

	on_doctype (a_name: STRING; an_id: XM_DTD_EXTERNAL_ID; has_internal_subset: BOOLEAN) is
			-- Document type declaration.
		local
			root: DEFINED_ELEMENT
		do
			root := locate_element (a_name)
			if root = Void then
				element_not_defined (Void, a_name)
			end
			state.set_doctype_root (root)
			precursor (a_name, an_id, has_internal_subset)
		end


feature {NONE} -- Tag

	on_attribute (a_namespace: STRING; a_prefix: STRING; a_local_part: STRING; a_value: STRING) is
			-- Process attribute.
		local
			attribute_name: STRING
			attribute: DEFINED_ATTRIBUTE
		do
			-- Skip this when no longer valid.
			if is_valid then
				attribute_name := resolved_name (a_namespace, a_local_part)
				attribute := state.current_element.get_attribute (attribute_name)
				if attribute = Void then
					attribute_not_defined (a_prefix, a_local_part)
				else
					if not attribute.is_valid_content (a_value) then
						attribute_content_invalid (a_prefix, a_local_part, a_value)
					else
						precursor (a_namespace, a_prefix, a_local_part, a_value)
						state.current_element_state.add_seen_attribute (attribute)
					end
				end
			end
		end

	on_start_tag (a_namespace: STRING; a_prefix: STRING; a_local_part: STRING) is
			-- Start of start tag.
		local
			name: UC_UTF8_STRING
			next_element: DEFINED_ELEMENT
			element_state: ELEMENT_STATE
			invalid_root: BOOLEAN
		do
				check
					have_state: state /= Void
				end

			check_valid_content

			debug ("XML")
				print ("<")
				print (a_local_part)
				print (">%N")
			end

			name := resolved_name (a_namespace, a_local_part)
			next_element := locate_element (name)
			if next_element = Void then
				-- is not a DEFINED_ELEMENT, halt
				element_not_defined (a_prefix, a_local_part)
			else
				if state.root = Void then
					-- check against root type set with DOCTYPE instruction
					invalid_root :=
						state.doctype_root /= Void and then
						state.doctype_root /= next_element
					if invalid_root then
						first_element_not_doctype_root (name)
					end
					-- Even when root is not correct, we do this, because a
					-- single callback might occur in `on_end_tag'.
					-- And it's easier to have state there, then to check
					-- if there is no root.
					state.set_root (next_element)
					create element_state.make (state.root)
					state.parents.put (element_state)
				else
					if is_valid_child (next_element) then
							check
								transition_ok: state.current_element = next_element
							end
						precursor (a_namespace, a_prefix, a_local_part)
					else
						invalid_child (state.current_element, next_element)
					end
				end
			end
		end

	on_start_tag_finish is
			-- Check here if we got all rquired attributes.
		local
			seen_attributes: DS_HASH_TABLE [BOOLEAN, DEFINED_ATTRIBUTE]
			attributes: DS_LINEAR [DEFINED_ATTRIBUTE]
			found: BOOLEAN
		do
			-- Skip this routine, when no longer valid.
			if is_valid then
				seen_attributes := state.current_element_state.seen_attributes
				attributes := state.current_element.attributes
				from
					attributes.start
					found := True
				until
					not found or else attributes.after
				loop
					if attributes.item_for_iteration.is_required then
						found := seen_attributes.has (attributes.item_for_iteration)
					end
					if found then
						attributes.forth
					end
				end
				if not found then
					required_attribute_not_found (attributes.item_for_iteration)
				else
					precursor
				end
			end
		end

	on_end_tag (a_namespace: STRING; a_prefix: STRING; a_local_part: STRING) is
			-- End tag.
		local
			name: UC_UTF8_STRING
			element: DEFINED_ELEMENT
		do
				check
					have_state: state /= Void
				end

			-- also check here if we got all required content...
			check_valid_content

			debug ("XML")
				print ("</")
				print (a_local_part)
				print (">%N")
			end

			name := resolved_name (a_namespace, a_local_part)
			-- this is usually catched by the parser, but in case it
			-- isn't...
			element := locate_element (name)
			if element = Void then
				element_not_defined (a_prefix, a_local_part)
			elseif state.current_element /= element then
				unexpected_end_tag (state.current_element, element)
			elseif is_valid_end (element) then
				precursor (a_namespace, a_prefix, a_local_part)
				state.parents.remove
			else
				required_child_expected (element)
			end
		end


feature -- Errors

	on_error (a_message: STRING) is
			-- Event producer detected an error. Probably a name space
			-- didn't resolve.
		do
			is_valid := False
		end


feature -- Content

	on_content (a_content: STRING) is
			-- Text content, just cached.
		do
			cached_content.append_string (a_content)
			-- do check on next event call, i.e. a start or stop tag.
		end


feature -- Where our events come from

	feed: XM_PARSER
			-- `feed' that passes the events that are considered.


feature -- Building schema definition

	add_element (element: DEFINED_ELEMENT) is
			-- Fill `elements'.
		require
			valid_element: element /= Void
			element_not_yet_added: not is_known_element (element.name)
		local
			key: UC_UTF8_STRING
		do
			-- Make sure key is an UC_UTF8_STRING, else hash_code differs
			-- between a plain STRING and a Unicode one.
			key ?= element.name
			if key = Void then
				create key.make_from_string (element.name)
			end
			elements.put (element, key)
		ensure
			element_added: is_known_element (element.name)
		end

	is_known_element (name: STRING): BOOLEAN is
			-- Is an element with `name' already added by `add_element'?
		local
			key: UC_UTF8_STRING
		do
			key ?= name
			if key = Void then
				create key.make_from_string (name)
			end
			Result := elements.has (key)
		end

	wipe_out_schema is
			-- Wipe out the schema.
		do
			elements.wipe_out
		ensure
			elements_clean: elements.count = 0
		end


feature {NONE} -- Schema definition state

	elements: DS_HASH_TABLE [DEFINED_ELEMENT, UC_UTF8_STRING]
			-- There is no name space support here, perhaps put every element
			-- in its own name space to start supporting it?
			-- Or just put the names with their expansion in elements?

	locate_element (name: STRING): DEFINED_ELEMENT is
		require
			valid_name: name /= Void and not name.is_empty
		local
			key: UC_UTF8_STRING
		do
			key ?= name
			if key = Void then
				create key.make_from_string (name)
			end
			elements.search (key)
			if elements.found then
				Result := elements.found_item
			end
		end


feature {NONE} -- State transition checks

	expected_element: DEFINED_ELEMENT is
			-- Currently expected element, can be Void.
		do
			Result := state.current_element_state.expected_element
		end

	is_valid_child (next_element: DEFINED_ELEMENT): BOOLEAN is
			-- Returns True if next_element is a valid child of
			-- `current_element', regarding all previous childs that we've
			-- seen.
		require
			next_element_not_void: next_element /= Void
			valid_current: state.current_element /= Void
			there_is_a_parent: not state.parents.is_empty
		local
			element_state: ELEMENT_STATE
		do
			Result := state.current_element.is_valid_child (state.current_element_state, next_element)
			if Result then
				-- child says yes, this tag becomes the current one
				create element_state.make (next_element)
				state.parents.put (element_state)
			end
		ensure
			parent_stack_maintained: Result implies state.current_element = next_element
		end

	is_valid_end (element: DEFINED_ELEMENT): BOOLEAN is
			-- Can `element' end this state?
		require
			element_not_void: element /= Void
			is_current_element: state.current_element = element
		do
			Result := state.current_element.is_valid_end (state.current_element_state)
		end


feature {NONE} -- Content validation

	cached_content: STRING
			-- We cache content (i.e. lump text nodes together), before
			-- doing the on_content call.

	check_valid_content is
			-- The actual on_content call is delayed. We wait until all
			-- text nodes have been gathered.
		do
			if not is_ignorable_content (cached_content) then
				if state.current_element.is_data_allowed then
					if state.current_element.is_valid_data (cached_content) then
						-- implement the forwarding myself, calling an
						-- inherited feature isn't very easy it seems.
						callbacks.on_content (cached_content)
					else
						element_does_not_allow_this_data (state.current_element, cached_content)
					end
				else
					element_does_not_allow_data (state.current_element)
				end
			end
			cached_content.wipe_out
		ensure
			cached_content_cleared: cached_content.is_empty
		end


feature -- XML didn't validate, what to do?

	attribute_not_defined (attribute_prefix, attribute_local_part: STRING) is
		do
			if attribute_prefix = Void then
				validation_error (format ("Tag ?s does not allow attribute '?s'.", <<current_path, attribute_local_part>>))
			else
				validation_error (format ("Tag ?s does not allow attribute '?s:?s'.", <<current_path, attribute_prefix.out, attribute_local_part.out>>))
			end
		end

	attribute_content_invalid (attribute_prefix, attribute_local_part, value: STRING) is
		do
			if attribute_prefix = Void then
				validation_error (format ("Content of attribute <?s ?s=%"?s%"> is not correct at path ?s.", <<state.current_element.name, attribute_local_part, value, current_path>>))
			else
				validation_error (format ("Content of attribute <?s ?s:?s=%"?s%"> is not correct at path ?s.", <<state.current_element.name, attribute_prefix.out, attribute_local_part.out, value, current_path>>))
			end
		end

	element_does_not_allow_data (element: DEFINED_ELEMENT) is
			-- Element does not allow content at all.
		do
			validation_error (format ("Element <?s> does not allow character data.", <<element.name.out>>))
		end

	element_not_defined (a_prefix, a_local_part: STRING) is
		do
			if a_prefix = Void then
				validation_error (format ("Tag <?s> is undefined.", <<a_local_part>>))
			else
				validation_error (format ("Tag <?s:?s> is undefined.", <<a_prefix, a_local_part>>))
			end
		end

	element_does_not_allow_this_data (element: DEFINED_ELEMENT; a_content: STRING) is
			-- Element rejected this content.
		do
			validation_error (format ("Element <?s> does not allow content '?s'.", <<element.name.out, a_content>>))
		end

	first_element_not_doctype_root (tag_name: STRING) is
		do
			validation_error (format ("First tag <?s> is not equal to the expected root.", <<tag_name.out>>))
		end


	invalid_child (parent, child: DEFINED_ELEMENT) is
		do
			if expected_element = Void then
				validation_error (format ("Parent <?s> does not allow child <?s> at the current position.", <<parent.name.out, child.name.out>>))
			else
				validation_error (format ("Parent <?s> expects <?s> at the current position, but got <?s>.", <<parent.name.out, expected_element.name.out, child.name.out>>))
			end
		end

	required_attribute_not_found (attribute: DEFINED_ATTRIBUTE) is
		do
			validation_error (format ("Required attribute '?s' not found in path ?s.", <<attribute.name, current_path>>))

		end

	unexpected_end_tag (parent, element: DEFINED_ELEMENT) is
		do
			validation_error (format ("Element </?s> cannot end tag <?s>.", <<parent.name.out, element.name.out>>))
		end

	required_child_expected (element: DEFINED_ELEMENT) is
		do
			if expected_element = Void then
				validation_error (format ("Tag </?s> comes too soon. Another tag was expected.", <<element.name.out>>))
			else
				validation_error (format ("Tag </?s> comes too soon. Tag <?s> expected.", <<element.name.out, expected_element.name.out >>))
			end
		end

	validation_error (error_msg: STRING) is
			-- Default strategy is to end parsing.
		do
			if is_valid then
				-- It seems we get here multiple times, have some trouble
				-- stopping parsing NOW if error occurs,
				-- i.e. set_end_of_document does not immediately stop.
				stop_parsing
				is_valid := False
				validation_error_extended_description := error_msg
				validation_error_position := position
			end
		end


feature -- Accessible parsing state

	current_element: DEFINED_ELEMENT is
		require
			parsing_started: is_inside_tag
		do
			Result := state.current_element
		ensure
			current_element_not_void: Result /= Void
		end

	current_tag_name: STRING is
			-- The name of the tag inside which the callbacks occur
		require
			inside_a_tag: is_inside_tag
		do
			Result := state.parents.item.element.name
		end

	current_path: STRING is
			-- Path in parsing, the concatenation of the name of all
			-- elements up tot the root.
		local
			top_item: ELEMENT_STATE
		do
			-- Recursively remove items from the stack and put them back
			-- on.
			if state.parents.is_empty then
				create {UC_UTF8_STRING} Result.make_empty
				Result.append_character ('/')
			else
				top_item := state.parents.item
				state.parents.remove
				Result := current_path
				if Result.item (Result.count) /= '/' then
					Result.append_character ('/')
				end
				Result.append_string (top_item.element.name)
				state.parents.put (top_item)
			end
		end

	is_inside_tag: BOOLEAN is
			-- Are callbacks occurring inside a tag?
			-- Before parsing the root tag, this is False.
		do
			Result := not state.parents.is_empty
		end


feature {NONE} -- Validation state

	is_ignorable_content (a_content: STRING): BOOLEAN is
			-- Consists `a_content' of white space and/or new lines?
		local
			i: INTEGER
			c: CHARACTER
		do
			from
				i := 1
				Result := True
			until
				not Result or else
				i > a_content.count
			loop
				c := a_content.item (i)
				Result :=
					c = ' ' or else
					c = '%N'  or else
					c = '%T' or else
					c = '%R'
				i:= i + 1
			end
		end

	is_valid: BOOLEAN

	state: PARSE_STATE

	validation_error_extended_description: STRING

	validation_error_position: XM_POSITION


feature {NONE} -- Implementation

	namespace_resolver: 	XM_NAMESPACE_RESOLVER
			-- Object that makes sure name spaces are resolved before we see them.

	stop_parsing is
			-- Try to stop parsing now.
		local
			xp: XM_EXPAT_PARSER
		do
			xp ?= feed
			if xp /= Void then
				xp.unregister_all_callbacks
			end
		end

feature {NONE} -- Once strinsg

	xmlns: STRING is "xmlns"


invariant

	feed_not_void: feed /= Void
	elements_not_void: elements /= Void
	cached_content_not_void: cached_content /= Void

end
